//
// Use, distribution and reproduction of this source is governed by
// the GNU Public License 2
//
// (c) 2000 Joshy (Jose Mejuto) jmejuto@pobox.com // joshy@cdngo.com
//
// 2000.08.21 First release to be added to the ogg vorbis project
//

// This file is intented to parse WAV headers from a file and report
// information to a main code portion without loosing any data if the
// file is not a valid WAV file, or the WAV file is damaged.
// stdin should work ok, but probably this code need a little tweak in
// the chunks parser to avoid memory problems in large padded WAV files
// to remove silences because the RollBackBuffer could grow to a undesirable
// limit.
// This piece of code should be easy to port from Win platforms to any other
// and should be endian free, but I can not test this point.
// The "int main()" function is included as an example on how use it.

#include <stdio.h>
#include <malloc.h>
#include <memory.h>

#include "WAVParser.h"

int LSBFirst=0;
int MSBFirst=0;

//Realloc memory in pages of 256 bytes, usually only one page is necessary, but with
//padded WAV files (To remove digital silences without rewritting the whole WAV file)
//the functions should perform 2 or 3 reallocs.
#define WAV_READ_PAGESIZE 256


//WAV header size limit could be almost (2^31)-9 bytes but for a safe process we can
//set a WAV limit near RAM memory, if the RollBackBuffer grows too much the function
//that read the WAV headers will return an error. If the parser detects a damaged
//WAV file it will return an error anyway. Usually leave this #define to 0x7FFFFFFF
//to avoid WAV header limits.
#define WAV_READ_MEMORYLIMIT 0x7FFFFFFF

int main(int argc, char *argv[])
{
    FILE *f;
    struct RIFF wavhdr;
    unsigned char *RollBackBuffer=NULL;
    int RollBackSize;

    // Test for endian and store in 
    // "MSBFirst" and "LSBFirst" global variables
    EndianMachine();

    // Open the file

    f=fopen(argv[1],"rb");
    if (!f) {
        printf("Where is the input file ???\n");
        return -1;
    }

    RollBackSize=0;
    RollBackBuffer=NULL;
    if (!Get_WAV_Header(f,&wavhdr,&RollBackBuffer,&RollBackSize)) {
        printf("Is not a WAV file.\n");
        //Free RollackBuffer if necessary
        if (RollBackBuffer) {
            free (RollBackBuffer);
            RollBackSize=0;
        }
        return -1;
    }
    //Test values useful for vorbize, this means 44100, 16 bits, stereo, PCM
    printf("\n");
    if (wavhdr.WAVE.fmt.FormatTag==1 && wavhdr.WAVE.fmt.Channels==2 &&
        wavhdr.WAVE.fmt.SamplesPerSec==44100 && wavhdr.WAVE.fmt.BitsPerSample==16) {
        printf("OK, this file is valid for vorbize, the stream starts at %d and file pointer is positioned there, even using stdin.\n",RollBackSize);
        printf("Previous information is stored in RollBackBuffer and header info\nin the wavhdr struct.\n");
    } else {
        printf("Sorry this file is not valid for vorbize, it should be 44100, 16 bits, stereo, PCM\n");
    }
    printf("\n");
    printf("FormatTag   =%d\n",(int)wavhdr.WAVE.fmt.FormatTag);
    printf("Channels    =%d\n",(int)wavhdr.WAVE.fmt.Channels);
    printf("Sampling    =%d\n",(int)wavhdr.WAVE.fmt.SamplesPerSec);
    printf("BitPerSample=%d\n",(int)wavhdr.WAVE.fmt.BitsPerSample);
    printf("Stream len  =%d\n",(int)wavhdr.WAVE.data.datalen);

    //Add this in the vorbize code
    //filelen=wavhdr.WAVE.data.datalen;

    if (RollBackBuffer) {
        free (RollBackBuffer);
        RollBackSize=0;
    }
    fclose (f);
    return 0;
}

int EndianMachine(void)
{
    int v;
    unsigned char *c;

    v=0x99553300;
    c=(unsigned char *)&v;
    if (*c==0) {
        LSBFirst=1;
        MSBFirst=0;
        return 0;
    } else {
        LSBFirst=0;
        MSBFirst=1;
        return 1;
    }
}

int Get_WAV_Header(FILE *f,struct RIFF *w,unsigned char **RollBackBuffer,int *RollBackSize)
{
    unsigned char *Buffer;
    int *PValue;        //Pointer to 32 bit value signed
    int Value;          //32 bit value signed
    int read;           //Read bytes in each action
    int CurrentPos;     //Current position for reads in RollBackBuffer
    int RollBackMax=0;  //Limit of allocated memory in RollBackBuffer
    int PadReadSize;    //For chunk information reads
    word *WordRead;     //Pointer to 16 bits value signed
    int LoopReady=0;    //Determine if the code reach the data chunk

    //RollBackBuffer should be an initialized pointer using malloc or best
    //initialized to NULL

    RollBackMax=*RollBackSize;
    CurrentPos=*RollBackSize;
    read=WAV_ReadFromStreamToRollBackBuffer(f,4,RollBackBuffer,RollBackSize,&RollBackMax);
    if (!read) {
        *RollBackBuffer=realloc(*RollBackBuffer,*RollBackSize);
        return 0;
    }
    Buffer=(char *)((*RollBackBuffer)+CurrentPos);
    // Test for RIFF header.
    if (memcmp(Buffer,"RIFF",4)==0) {
        //Fill WAVE struct field
        memcpy(w->riffheader,Buffer,4);
        CurrentPos=CurrentPos+read;
        // Now Test for RIFF chunk size
        read=WAV_ReadFromStreamToRollBackBuffer(f,4,RollBackBuffer,RollBackSize,&RollBackMax);
        if (!read) {
            *RollBackBuffer=realloc(*RollBackBuffer,*RollBackSize);
            return 0;
        }
        PValue=(int *)((*RollBackBuffer)+CurrentPos);
        Value=GetLSBFirst32Bits(*PValue);
        //Fill WAVE WAVELen field
        w->WAVElen=Value;
        
        // Now test for WAVE RIFF chunk
        CurrentPos=CurrentPos+read;
        // Now Test for RIFF chunk size
        read=WAV_ReadFromStreamToRollBackBuffer(f,4,RollBackBuffer,RollBackSize,&RollBackMax);
        if (!read) {
            *RollBackBuffer=realloc(*RollBackBuffer,*RollBackSize);
            return 0;
        }
        Buffer=(char *)((*RollBackBuffer)+CurrentPos);
        if (memcmp(Buffer,"WAVE",4)==0) {
            //OK, WAVE chunk is present
            //This is not a good format in vorbis code, only copy the
            //read part
            memcpy(w->WAVE.fmtheader,Buffer,4);
            CurrentPos=CurrentPos+read;
            //Enter the loop to read all chunks before DATA chunk
            LoopReady=0;
            while (!LoopReady) {
                //Read chunk name
                read=WAV_ReadFromStreamToRollBackBuffer(f,8,RollBackBuffer,RollBackSize,&RollBackMax);
                if (!read) {
                    *RollBackBuffer=realloc(*RollBackBuffer,*RollBackSize);
                    return 0;
                }
                //Transform it in a 32bit LSB value
                PValue=(int *)((*RollBackBuffer)+CurrentPos);
                Value=GetLSBFirst32Bits(*PValue);
                //Read the chunk size
                PValue=(int *)((*RollBackBuffer)+CurrentPos+4);
                PadReadSize=GetLSBFirst32Bits(*PValue);
                CurrentPos=CurrentPos+read;
                switch (Value) {
                case 0x20746d66: //"fmt " chunk
                    //Copy fmt chunk to vorbize RIFF struct
                    memcpy(&w->WAVE.fmtheader[4],"fmt ",4);
                    w->WAVE.fmtlen=PadReadSize;

                    if (PadReadSize<14) {
                        //Damaged WAV file, it should be at least 14 bytes long
                        *RollBackBuffer=realloc(*RollBackBuffer,*RollBackSize);
                        return 0;
                    }

                    //Read 14 bytes fmt;
                    read=WAV_ReadFromStreamToRollBackBuffer(f,PadReadSize,RollBackBuffer,RollBackSize,&RollBackMax);
                    if (!read) {
                        *RollBackBuffer=realloc(*RollBackBuffer,*RollBackSize);
                        return 0;
                    }
                    Buffer=(char *)((*RollBackBuffer)+CurrentPos);
                    memcpy (&w->WAVE.fmt,Buffer,14);

                    //Solve endian possible problems
                    w->WAVE.fmt.FormatTag=GetLSBFirst16Bits(w->WAVE.fmt.FormatTag);
                    w->WAVE.fmt.Channels=GetLSBFirst16Bits(w->WAVE.fmt.Channels);
                    w->WAVE.fmt.SamplesPerSec=GetLSBFirst32Bits(w->WAVE.fmt.SamplesPerSec);
                    w->WAVE.fmt.AvgBytesPerSec=GetLSBFirst32Bits(w->WAVE.fmt.AvgBytesPerSec);
                    w->WAVE.fmt.BlockAlign=GetLSBFirst16Bits(w->WAVE.fmt.BlockAlign);
                    if (w->WAVE.fmt.FormatTag==1) { //PCM WAV file
                        if (PadReadSize >= 16) {
                            WordRead=(word *)((*RollBackBuffer)+CurrentPos+14);
                            w->WAVE.fmt.BitsPerSample=GetLSBFirst16Bits(*WordRead);
                        } else {
                            //Invalid WAV PCM File
                            *RollBackBuffer=realloc(*RollBackBuffer,*RollBackSize);
                            return 0;
                        }
                    }
                    break;
                case 0x61746164:
                    //data chunk
                    memcpy(w->WAVE.data.dataheader,"data",4);
                    w->WAVE.data.datalen=PadReadSize;
                    LoopReady=1;
                    read=0; //No further reads in this chunk
                    break;
                default:
                    //Read the chunk information to the RollBackBuffer
                    read=WAV_ReadFromStreamToRollBackBuffer(f,PadReadSize,RollBackBuffer,RollBackSize,&RollBackMax);
                    if (!read) {
                        *RollBackBuffer=realloc(*RollBackBuffer,*RollBackSize);
                        return 0;
                    }
                    break;
                }
                CurrentPos=CurrentPos+read;
            }
        }
    }
    *RollBackBuffer=realloc(*RollBackBuffer,*RollBackSize);
    return LoopReady; //=1 if we read all chuncks before data chunk
}

int GetLSBFirst32Bits(int Value)
{
    unsigned char *buf;
    if (LSBFirst) {
        return Value;
    } else {
        buf=(unsigned char *)&Value;
        return (buf[0]<<24) + (buf[1]<<16) + (buf[2]<<8) + buf[3];
    }
}

short int GetLSBFirst16Bits(int Value)
{
    unsigned char *buf;
    if (LSBFirst) {
        return Value;
    } else {
        buf=(unsigned char *)&Value;
        return (buf[2]<<8) + buf[3];
    }
}

int WAV_ReadFromStreamToRollBackBuffer(FILE *f, int bytes,unsigned char **RollBackBuffer,int *RollBackSize,int *RollBackMax)
{
    int read;
    int AllocSize=0;

    if (*RollBackSize+bytes>WAV_READ_MEMORYLIMIT) {
        //WAV header greater than WAV_READ_MEMORYLIMIT returning error.
        return 0;
    }

    if (*RollBackSize+bytes >= *RollBackMax) {
        // Allocate necesary bytes
        AllocSize=*RollBackSize+bytes-*RollBackMax;
        if (AllocSize<WAV_READ_PAGESIZE) {
            AllocSize=WAV_READ_PAGESIZE;
        }
        *RollBackBuffer=(unsigned char *)realloc(*RollBackBuffer,(*RollBackMax+AllocSize));
        if (!*RollBackBuffer) {
            // No memory, reset variables
            *RollBackSize=0;
            *RollBackMax=0;
            return 0;
        }
        *RollBackMax=*RollBackMax+AllocSize;
    }
    read=fread((*RollBackBuffer)+*RollBackSize,1,bytes,f);
    *RollBackSize=*RollBackSize+read;
    if (read<bytes) {
        return 0;
    } else {
        return read;
    }
}

